

class PatchType(object):

    MergePatchType = "application/merge-patch+json"
    JSONPatchType = "application/json-patch+json"
    StrategicMergePatchType = "application/strategic-merge-patch+json"
    ApplyPatchType =  "application/apply-patch+yaml"